/*
 *      Winbond W83627HF Watchdog Timer driver for Linux 2.2.x and 2.4.x
 *
 *      This code is based on wdt.c with original copyright
 *
 *	(c) Copyright 1996-1997 Alan Cox <alan@redhat.com>, All Rights Reserved.
 *				http://www.redhat.com
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	as published by the Free Software Foundation; either version
 *	2 of the License, or (at your option) any later version.
 *	
 *	Neither Alan Cox, Matthias Cramer nor Dolphins Network Systems 
 *      admit liability nor provide warranty for any of this software. 
 *      This material is provided "AS-IS" and at no charge.	
 *
 *	(c) Copyright 2000    Matthias Cramer <cramer@dolphins.ch>
 *
 *	Release 0.01.
 *
 */

/*

Compile comand: "gcc -D__KERNEL__ -DMODULE -Ilinux/include -Wall -Wstrict-prototypes -O2 -m486 -c w83627hf-wdt.c"

*/


#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/watchdog.h>
#include "w83627hf-wdt.h"
#include <linux/malloc.h>
#include <linux/ioport.h>
#include <linux/fcntl.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/notifier.h>
#include <linux/reboot.h>
#include <linux/init.h>

static int w83627_is_open=0;

/*
 *	You must set these - there is no sane way to probe for this board.
 *	You can use w83627=x to set these now.
 */
 
static int io=0x2E;			/* Normany used addres on Portwell Boards */

#define WD_TIMO (3)		/* 5 minutes (1 ... 255) */

#ifndef MODULE

/**
 *	w83627_setup:
 *	@str: command line string
 *
 *	Setup options. The board isn't really probe-able so we have to
 *	get the user to tell us the configuration. Sane people build it 
 *	modular but the others come here.
 */
 
static int __init w83627_setup(char *str)
{
	int ints[4];

	str = get_options (str, ARRAY_SIZE(ints), ints);

	if (ints[0] > 0)
	{
		io = ints[1];
	}

	return 1;
}

__setup("w83627=", w83627_setup);

#endif /* !MODULE */
 
/*
 *	Programming support
 */
 
static void w83627_enter_ext_mode()
{
 	outb_p(0x87, WDT_EFER);	/* Enter extended function mode */
 	outb_p(0x87, WDT_EFER);	/* Do it twice !! */
 	
}

static void w83627_leave_ext_mode()
{
 	outb_p(0x0AA, WDT_EFER);	/* Leave extended function mode */
}

static void w83627_configure(char w83627_time)
{
	char	cr_data=0;
	
	w83627_enter_ext_mode();
	outb_p(0x2B, WDT_EFER);		/* Select CR2B (GPIO multiplexed pin selection register 2. VCC powered. Default 0XC0) */
	cr_data = inb (WDT_EFDR);	/* Read the data from CR2B */

	#ifdef DEBUG
        printk(KERN_INFO "W83627 - Configure (%d), CR2B 0x%x -> ", w83627_time, cr_data);
        #endif

	cr_data &= 0xEF;		/* Set bit 4 in CR2B to 0 to select watch dog timer output */
	cr_data &= 0xFF;

	#ifdef DEBUG
        printk(KERN_INFO "... 0x%x\n", cr_data);
        #endif

	outb_p(cr_data, WDT_EFDR);	/* Write back the new value */
	
	outb_p(0x08, WDT_EFER);		/* Select locical device 8 */
	outb_p(0x0F5, WDT_EFER);		/* Select config register F5 */
	cr_data = inb(WDT_EFDR);	/* Read CRF5 */

	#ifdef DEBUG
        printk(KERN_INFO "W83627 - Configure (%d), CRF5 0x%x -> ", w83627_time, cr_data);
        #endif

	cr_data |= 0x08;		/* Set bit 3 to 1 to select the minute counter */

	#ifdef DEBUG
        printk(KERN_INFO "... 0x%x\n", cr_data);
        #endif

	outb_p(0x5F, WDT_EFER);		/* Select CRF5 */
	outb_p(cr_data, WDT_EFDR);	/* Write new value to CRF5) */
	
	outb_p(0x0F6, WDT_EFER);		/* Select CRF6 */
	outb_p(w83627_time, WDT_EFDR);	/* Write Timeout counter to CRF6 */
	
	w83627_leave_ext_mode();
	#ifdef DEBUG
        printk(KERN_INFO "W83627 - Configure (%d)\n", w83627_time);
        #endif
}


/*
 *	Kernel methods.
 */
 
 
/**
 *	w83627_status:
 *	
 *	Extract the status information from a WDT watchdog device. There are
 *	several board variants so we have to know which bits are valid. Some
 *	bits default to one and some to zero in order to be maximally painful.
 *
 *	we then map the bits onto the status ioctl flags.
 */
 
static int w83627_status(void)
{
	/* Not supported */
		
	return 1;
}

static long long w83627_llseek(struct file *file, long long offset, int origin)
{
	return -ESPIPE;
}

/**
 *	w83627_write:
 *	@file: file handle to the watchdog
 *	@buf: buffer to write (unused as data does not matter here 
 *	@count: count of bytes
 *	@ppos: pointer to the position to write. No seeks allowed
 *
 *	A write to a watchdog device is defined as a keepalive signal. Any
 *	write of data will do, as we we don't define content meaning.
 */
 
static ssize_t w83627_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	/*  Can't seek (pwrite) on this device  */
	if (ppos != &file->f_pos)
		return -ESPIPE;

	if(count)
	{
		w83627_configure(WD_TIMO);
		return 1;
		#ifdef DEBUG
		printk(KERN_INFO "W83627 - Write\n");
		#endif
	}
	return 0;
}


/*
 *      Read reports the temperature in degrees Fahrenheit.
 */
static ssize_t w83627_read(struct file *file, char *buf, size_t count, loff_t *ptr)
{

	char timeout;
	
	w83627_enter_ext_mode();
	outb_p(0x08, WDT_EFER);		/* Select locical device 8 */
	outb_p(0x0F6, WDT_EFER);		/* Select CRF6 */
	timeout = inb(WDT_EFDR);	/* Read Timeout counter from CRF6 */
	
	w83627_leave_ext_mode();
	
	if(copy_to_user(buf,&timeout,1))
		return -EFAULT;
	return 1;
}

 /**
 *	w83627_ioctl:
 *	@inode: inode of the device
 *	@file: file handle to the device
 *	@cmd: watchdog command
 *	@arg: argument pointer
 *
 *	The watchdog API defines a common set of functions for all watchdogs
 *	according to their available features. We only actually usefully support
 *	querying capabilities and current status. 
 */
 
static int w83627_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	unsigned long arg)
{
	static struct watchdog_info ident=
	{
		1,
		1,
		"W83626(HF/F)"
	};
	
	ident.options=1;	/* Mask down to the card we have */
	switch(cmd)
	{
		default:
			return -ENOIOCTLCMD;
		case WDIOC_GETSUPPORT:
			return copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident))?-EFAULT:0;

		case WDIOC_GETSTATUS:
			return put_user(w83627_status(),(int *)arg);
		case WDIOC_GETBOOTSTATUS:
			return put_user(0, (int *)arg);
		case WDIOC_KEEPALIVE:
			w83627_configure(WD_TIMO);
			return 0;
	}
}

/**
 *	w83627_open:
 *	@inode: inode of device
 *	@file: file handle to device
 *
 *	One of our two misc devices has been opened. The watchdog device is
 *	single open and on opening we load the counters. Counter zero is a 
 *	100Hz cascade, into counter 1 which downcounts to reboot. When the
 *	counter triggers counter 2 downcounts the length of the reset pulse
 *	which set set to be as long as possible. 
 */
 
static int w83627_open(struct inode *inode, struct file *file)
{
	switch(MINOR(inode->i_rdev))
	{
		case WATCHDOG_MINOR:
			if(w83627_is_open)
				return -EBUSY;
			/*
			 *	Activate 
			 */
	 
			w83627_is_open=1;
			w83627_configure(0);
			w83627_configure(WD_TIMO);
			#ifdef DEBUG
			printk(KERN_INFO "W83627 - Open\n");
			#endif
			return 0;
		case TEMP_MINOR:
			return 0;
		default:
			return -ENODEV;
	}
}

/**
 *	w83627_close:
 *	@inode: inode to board
 *	@file: file handle to board
 *
 *	The watchdog has a configurable API. There is a religious dispute 
 *	between people who want their watchdog to be able to shut down and 
 *	those who want to be sure if the watchdog manager dies the machine
 *	reboots. In the former case we disable the counters, in the latter
 *	case you have to open it again very soon.
 */
 
static int w83627_release(struct inode *inode, struct file *file)
{
	if(MINOR(inode->i_rdev)==WATCHDOG_MINOR)
	{
#ifndef CONFIG_WATCHDOG_NOWAYOUT	
		w83627_configure(0);
#endif		
		w83627_is_open=0;
		#ifdef DEBUG
		printk(KERN_INFO "W83627 - Release\n");
		#endif
	}
	return 0;
}

/**
 *	notify_sys:
 *	@this: our notifier block
 *	@code: the event being reported
 *	@unused: unused
 *
 *	Our notifier is called on system shutdowns. We want to turn the card
 *	off at reboot otherwise the machine will reboot again during memory
 *	test or worse yet during the following fsck. This would suck, in fact
 *	trust me - if it happens it does suck.
 */

static int w83627_notify_sys(struct notifier_block *this, unsigned long code,
	void *unused)
{
	if(code==SYS_DOWN || code==SYS_HALT)
	{
		/* Turn the card off */
		w83627_configure(0);
	}
	return NOTIFY_DONE;
}
 
/*
 *	Kernel Interfaces
 */
 
 
static struct file_operations w83627_fops = {
#if ( LINUX_VERSION_CODE < 132096 )
	w83627_llseek,
        w83627_read,
        w83627_write,
        NULL,           /* No Readdir */
        NULL,           /* No Select */
        w83627_ioctl,
        NULL,           /* No mmap */
        w83627_open,
        NULL,           /* flush */
        w83627_release
#else
	owner:		THIS_MODULE,
	llseek:		w83627_llseek,
	read:		w83627_read,
	write:		w83627_write,
	ioctl:		w83627_ioctl,
	open:		w83627_open,
	release:	w83627_release,
#endif
};

static struct miscdevice w83627_miscdev=
{
	WATCHDOG_MINOR,
	"watchdog",
	&w83627_fops
};

/*
 *	The WDT card needs to learn about soft shutdowns in order to
 *	turn the timebomb registers off. 
 */
 
static struct notifier_block w83627_notifier=
{
	w83627_notify_sys,
	NULL,
	0
};

#ifdef MODULE

#define w83627_init init_module

/**
 *	cleanup_module:
 *
 *	Unload the watchdog. You cannot do this with any file handles open.
 *	If your watchdog is set to continue ticking on close and you unload
 *	it, well it keeps ticking. We won't get the interrupt but the board
 *	will not touch PC memory so all is fine. You just have to load a new
 *	module in 60 seconds or reboot.
 */
 
void cleanup_module(void)
{
	misc_deregister(&w83627_miscdev);
	unregister_reboot_notifier(&w83627_notifier);
/*	release_region(io,2); */
}

#endif

/**
 * 	w83627_init:
 *
 *	Set up the WDT watchdog board. All we have to do is grab the
 *	resources we require and bitch if anyone beat us to them.
 *	The open() function will actually kick the board off.
 */
 
int __init w83627_init(void)
{
	printk(KERN_INFO "W83627 WDT driver 0.01 at io 0x%X\n", WDT_EFER);

	misc_register(&w83627_miscdev);

/*	request_region(io, 2, "w83627");	*/
	register_reboot_notifier(&w83627_notifier);
	return 0;
}
